;; -* emacs-lisp *-
;; epoch.el:  Basic code setup and loading.
;;
;; Original Author: Alan M. Carroll
;; Extensively rewritten by Simon Kaplan
;; Ideas and code contributed by Denys Duchier
;; Reworked by Alan M. Carroll

;; epoch-version: 3.2-alpha

;; This file is part of Epoch, a modified version of GNU Emacs.

;; GNU Emacs is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 1, or (at your option)
;; any later version.

;; GNU Emacs is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.
;;; --------------------------------------------------------------------------
(require 'epoch-util)			;various utilities
(require 'event)
;; epoch stuff.  key binding, building on primitive routines, etc.

;; user-definable switches
;;;
;;; Installed events
(install-event 'map)
(install-event 'resize)
(install-event 'move)
(install-event 'client-message)
(install-event 'focus)
(install-event 'property-change)
(install-event 'button)
(install-event 'motion)

;; auto-raise on focus event.
(defvar auto-raise-screen t
"Sets focus events to automatically raise the screen")

;; include system name in screen titles
(defvar include-system-name t
"Sets the system name to be included in the screen title")

(defvar wm-protocol-list
  (list
    (epoch::intern-atom "WM_DELETE_WINDOW")
  )
  "List of protocol atoms that epoch recognizes"
)

(defvar xa-wm-protocols (epoch::intern-atom "WM_PROTOCOLS"))

;;; --------------------------------------------------------------------------
;; set the initial screen to have the right protocols
(epoch::set-property xa-wm-protocols wm-protocol-list 1)
;; need to map the screen too
(epoch::mapraised-screen 1)
;;
;;; --------------------------------------------------------------------------
;; epoch-mode-alist sets default screen creation params by mode

(defvar epoch-mode-alist
  (list
    (cons 'plain-tex-mode
      (list
	(cons 'geometry "80x52")
	(cons 'cursor-glyph 2)))
    (cons 'latex-mode
      (list
	(cons 'geometry "80x52")
	(cons 'cursor-glyph 2)
      )
    )
  )
)

;;; --------------------------------------------------------------------------
;; add update flag to the screen properties and set globally true,
;; true for first screen.

(setq epoch::global-update t)
;; must to set-update initially to make this true of first screen.
(dolist (s (epoch::screen-list t)) (epoch::set-update t s))
;;; set up for future screens
(push '(update t) epoch::screen-properties)
;;; --------------------------------------------------------------------------
;; restarts will be nil the first time you eval this file, t thereafter.
;; This prevents screens from getting real squirrly names like foo@bar@bar...

(defvar restarts nil)

;; if this is the first time we are evaling .emacs, and we are
;; running epoch, set the name of the first screen and the minibuffer to be
;; <buffer-name> @ <machine-name> if we are running with include-system-name
;; set to t, and just <buffer-name> otherwise.

(defun sys-name ()
  (if include-system-name
      (concat " @ " (system-name))
    ""))


(if (not restarts)
  (progn
    (epoch::title (concat (buffer-name (current-buffer)) (sys-name)))
    (epoch::title (concat (epoch::title nil 0) (sys-name)) 0)))

;; and now make sure we dont do that again.  This is all useful when 
;; debuffing stuff.

(setq restarts t)

;;; --------------------------------------------------------------------------
;; interfaces to prim routines.
(defvar *create-screen-alist-hook* nil
 "A hook that is called with the create-screen alist just before the screen is actually created. Should return a new alist (or the old one if no change).")
;;;
(defun create-screen (&optional buff alist)
"Create a new edit screen. Optional BUFFER indicates the buffer to put in\n\
root window of the screen. Optional ALIST is a list of attributes for the\n\
new screen. ALIST is augmented by searching epoch-mode-alist for mode\n\
dependent attributes."
  (interactive)
  (let*
    (
      (b (and buff (get-buffer buff)))
      (xname (or (and b (buffer-name b)) "Edit"))
      (real-alist
	(append
	  alist
	  (get-epoch-mode-alist-contents buff)
	  (list (cons 'title xname) (cons 'icon-name xname))
	)
      )
      scr
    )
    ;; call the alist adjust hook
    (cond
      ((listp  *create-screen-alist-hook*)
	(dolist (hook *create-screen-alist-hook*)
	  (setq real-alist (funcall hook real-alist))
	)
      )
      ((functionp *create-screen-alist-hook*)
	(setq real-alist (funcall *create-screen-alist-hook* real-alist))
      )
    )
    ;; create the screen
    (setq scr (epoch::create-screen buff real-alist))
    ;; set up the buffers and what-not.
    (and buff (get-buffer buff)
      (save-excursion (set-buffer buff) (setq screens (cons scr screens)))
    )
    (epoch::set-property xa-wm-protocols wm-protocol-list scr)
    (if (not (cdr (assq 'no-map real-alist))) (epoch::mapraised-screen scr))
    scr
  )
)

(defun remove-screen-from-buffer (buf)
  (save-excursion
    (set-buffer buf)
    (setq screens (delq the-scr screens))))

(defun delete-screen (&optional scr)
  (interactive)
  (let ((the-scr (or scr (current-screen))))
    (mapcar (symbol-function 'remove-screen-from-buffer) (buffer-list)))
  (epoch::delete-screen scr)
;  (if (not (screen-mapped-p (current-screen))) (mapraised-screen))
)

(defun select-screen (&optional scr)
  (interactive)
  (epoch::raise-screen 0)
  (epoch::select-screen scr))

(defun unmap-screen (&optional scr)
  (interactive)
  (epoch::unmap-screen scr)
  (if (or (null scr) (eq (epoch::current-screen) scr) )
    (epoch::select-screen)
  )
  scr					;return a better value
)

(defun iconify-screen (&optional scr)
  (interactive)
  (epoch::iconify-screen scr)
  (if (or (null scr) (eq (epoch::current-screen) scr) )
    (epoch::select-screen)
  )
  scr					;return a better value
)

(defun cursor-to-screen (scr)
  (interactive)
  (if auto-raise-screen (raise-screen scr))
  (epoch::warp-mouse (/ (epoch::screen-width scr) 2) 0 scr)
)

(fset 'redisplay-screen (symbol-function 'epoch::redisplay-screen))
(fset 'set-screen-modified (symbol-function 'epoch::set-screen-modified))
(fset 'next-screen (symbol-function 'epoch::next-screen))
(fset 'prev-screen (symbol-function 'epoch::prev-screen))
(fset 'current-screen (symbol-function 'epoch::current-screen))
(fset 'minibuf-screen (symbol-function 'epoch::minibuf-screen))
(fset 'raise-screen (symbol-function 'epoch::raise-screen))
(fset 'lower-screen (symbol-function 'epoch::lower-screen))
(fset 'map-screen (symbol-function 'epoch::map-screen))
(fset 'background (symbol-function 'epoch::background))
(fset 'foreground (symbol-function 'epoch::foreground))
(fset 'cursor-color (symbol-function 'epoch::cursor-color))
(fset 'cursor-glyph (symbol-function 'epoch::cursor-glyph))
(fset 'set-update (symbol-function 'epoch::set-update))
(fset 'update-p (symbol-function 'epoch::update-p))
(fset 'icon-name (symbol-function 'epoch::icon-name))
(fset 'title (symbol-function 'epoch::title))
(fset 'screenp (symbol-function 'epoch::screen-p))
(fset 'screen-p (symbol-function 'epoch::screen-p))
(fset 'screen-of-window (symbol-function 'epoch::screen-of-window))
(fset 'screens-of-buffer (symbol-function 'epoch::screens-of-buffer))
(fset 'get-screen (symbol-function 'epoch::get-screen))
(fset 'get-screen-id (symbol-function 'epoch::get-screen-id))
; for compatibility (next 2)
(fset 'epoch::make-id (symbol-function 'epoch::get-screen-id))
(fset 'epoch::make-screen (symbol-function 'epoch::get-screen))
;
(fset 'font (symbol-function 'epoch::font))
(fset 'query-mouse (symbol-function 'epoch::query-mouse))
(fset 'query-pointer (symbol-function 'epoch::query-pointer))
(fset 'query-cursor (symbol-function 'epoch::query-cursor))
(fset 'warp-mouse (symbol-function 'epoch::warp-mouse))
(fset 'warp-pointer (symbol-function 'epoch::warp-pointer))
(fset 'screen-list (symbol-function 'epoch::screen-list))
(fset 'move-screen (symbol-function 'epoch::move-screen))
(fset 'mapraised-screen (symbol-function 'epoch::mapraised-screen))
(fset 'screen-information (symbol-function 'epoch::screen-information))
(fset 'plane-size (symbol-function 'epoch::plane-size))
(fset 'screen-mapped-p (symbol-function 'epoch::screen-mapped-p))
(fset 'change-screen-size (symbol-function 'epoch::change-screen-size))

(fset 'intern-atom (symbol-function 'epoch::intern-atom))
(fset 'unintern-atom (symbol-function 'epoch::unintern-atom))
(fset 'resourcep (symbol-function 'epoch::resourcep))
(fset 'string-to-resource (symbol-function 'epoch::string-to-resource))
(fset 'resource-to-string (symbol-function 'epoch::resource-to-string))
(fset 'resource-to-type (symbol-function 'epoch::resource-to-type))
(fset 'set-resource-type (symbol-function 'epoch::set-resource-type))
(fset 'get-property (symbol-function 'epoch::get-property))
(fset 'set-property (symbol-function 'epoch::set-property))
(fset 'send-client-message (symbol-function 'epoch::send-client-message))
(fset 'xid-of-screen (symbol-function 'epoch::xid-of-screen))
(fset 'query-tree (symbol-function 'epoch::query-tree))
(fset 'rebind-key (symbol-function 'epoch::rebind-key))
(fset 'mod-to-shiftmask (symbol-function 'epoch::mod-to-shiftmask))

(fset 'flash-screen (symbol-function 'epoch::flash-screen))
(fset 'set-bell (symbol-function 'epoch::set-bell))
;;; --------------------------------------------------------------------------
(defun find-buffer-other-screen (buffer)
  "Switch to BUFFER in other screen.  If buffer is already in another screen then select that, else make a new screen."
  (interactive "BSwitch to buffer other screen: ")
  (setq target-buffer (get-buffer buffer))
  (when (not target-buffer)
    (setq target-buffer (get-buffer-create buffer))
    (save-excursion
      (set-buffer target-buffer)
      (setq screens nil)
    )
  )
  (let
    (
      (scr
	(car (delq (current-screen) (epoch::screens-of-buffer target-buffer)))
      )
      (xname (concat (buffer-name target-buffer) (sys-name)))
    )
    (when (null scr)
      (setq scr
	(create-screen
	  target-buffer
	  (list (cons 'title xname) (cons 'icon-name xname))
	)
      )
    )
    (if (screen-mapped-p scr)
      (cursor-to-screen (select-screen scr))
      (progn
	(on-map-do scr 'cursor-to-screen)
	(mapraised-screen (select-screen scr))
      )
    )
    (select-window (get-buffer-window target-buffer))
    target-buffer			;return value
  )
)

(defun find-file-other-screen (filename)
  "Find file in other screen"
  (interactive "FFind file other screen: ")
  (setq target-buffer (find-file-noselect filename))
  (if (bufferp target-buffer) (find-buffer-other-screen target-buffer))
)
;;; --------------------------------------------------------------------------
(defun switch-screen (&optional scr)
  "Switch to next screen, and move focus there.
   If called with optional argument, then goto that screen instead."
  (interactive)
  (select-screen (if scr scr (next-screen)))
  (if auto-raise-screen (raise-screen (current-screen) ))
  (cursor-to-screen  (current-screen)))

(defun prev-switch-screen (&optional scr)
  "Switch to next screen, and move focus there.
   If called with optional argument, then goto that screen instead."
  (interactive)
  (select-screen (if scr scr (prev-screen)))
  (if auto-raise-screen (raise-screen (current-screen) ))
  (cursor-to-screen  (current-screen)))

(defun switch-screen-noselect ()
  "Switch to next screen, without altering the focus.  Used to allow circulation through screens without moving the mouse"
  (interactive)
  (select-screen (next-screen))
  (if auto-raise-screen (raise-screen (current-screen))))

(defun dired-other-screen (&optional dirname)
  "Pop up another screen and run dired in it"
  (interactive "DDired other screen (directory name) ")
  (let
    (
      (new-screen (select-screen (create-screen)))
    )
    (on-map-do new-screen (function (lambda (s) (cursor-to-screen s))))
    (kill-buffer (current-buffer))
    (dired dirname)))

(defun duplicate-screen ()
  "Makes a copy of current buffer in new screen"
  (interactive)
  (setq target-buffer (current-buffer))
  (select-screen 
   (create-screen target-buffer
		    (list (cons 'title (concat (buffer-name target-buffer )
						 (sys-name))))))
  (sit-for 2)
  (cursor-to-screen (current-screen))
  (sit-for 0))

 
(defun send-focus-to-current-screen ()
  "Focus on current screen"
  (interactive)
  (select-screen (current-screen))
  (if auto-raise-screen (raise-screen (current-screen) )))

(defun remove-screen (&optional scr)
  "Delete the argument screen, or current screen if nil.  
  Just an interactive interface to delete-screen"
  (interactive)
  (delete-screen (or scr (current-screen)))
  (focus-current-screen))

(defun focus-current-screen ()
  "Focus on and raise current screen"
  (interactive)
  (cursor-to-screen (current-screen))
  (if auto-raise-screen (raise-screen (current-screen))))

;; mode handling stuff.  The variable epoch-mode-alist contains mode/default
;; pairs in an alist structure.  This is then used by create-screen.  The
;; list so obtained is appended to the list passed into create-screen,
;; so as to allow easy override of the defaults.  The function 
;; get-epoch-mode-alist-contents does the work for us.
;; The variable epoch-mode-alist is defined up at the top of the file.

(defun get-epoch-mode-alist-contents (&optional buffer)
  "find the alist for mode of buffer.  if nil, use current buffer's mode"
  (let ( (buff (and buffer (get-buffer buffer))) )
    (if (bufferp buff)
      (save-excursion
	(set-buffer buffer)
	(assoc major-mode epoch-mode-alist)
      )
      (assoc major-mode epoch-mode-alist)
    )
  )
)
;;; --------------------------------------------------------------------------
;;; Event handling (see event.el for the underlying implementation)
;;; deal with focus event
;;;
;;; handle focus events
(push-event 'focus
  (function
    (lambda (type new-focus-state scr)
      (when new-focus-state
	(select-screen scr)
	(if auto-raise-screen (raise-screen scr))
      )
    )
  )
)
;;; --------------------------------------------------------------------------
;;; Handler that maintains on-event-do lists
(defvar on-event::do-list (list '(map) '(unmap) '(move) '(resize))
"Alist of event, screen/action lists"
)
;;;
(defun on-event::handler (type value screen)
"Handle events that have one-shot actions. Types are controlled by the
variable on-event::do-list, which is an alist of event types, with the
value an alist of screens and actions."
  (if (and (eq 'map type) (not value)) (setq type 'unmap))
  (let*
    (
      (epoch::event-handler-abort nil)	;no aborting!
      (the-list (assq type on-event::do-list))
      (info (assq screen (cdr the-list)))
      (action (cdr info))
    )
    (when info
      (setcdr the-list (alist-delete screen (cdr the-list)))
      (when (functionp action)
	(cond
	  ((or (eq type 'map) (eq type 'unmap))
	    (funcall action screen)
	  )
	  (t (funcall action screen value))
	)
      )
    )
  )
)
(push-event 'map 'on-event::handler)
(push-event 'move 'on-event::handler)
(push-event 'resize 'on-event::handler)
;;;
(defun on-map-do (screen action)
"On the next map event for SCREEN, call ACTION (a function of 1 argument,
the screen)"
  (let ( (mlist (assq 'map on-event::do-list)) )
    (push (cons (get-screen screen) action) (cdr mlist))
  )
)
(defun on-unmap-do (screen action)
"On the next unmap event for SCREEN, call ACTION (a function of 1 argument,
the screen)"
  (let ( (mlist (assq 'unmap on-event::do-list)) )
    (push (cons (get-screen screen) action) (cdr mlist))
  )
)
;;;
(defun on-resize-do (screen action)
"On the next resize event for SCREEN, call ACTION (a function of 2 arguments,
screen and event value)"
  (let ( (rlist (assq 'resize on-event::do-list)) )
    (push (cons (get-screen screen) action) (cdr rlist))
  )
)
;;;
(defun on-move-do (screen action)
"On the next move event for SCREEN, call ACTION (a function of 2 arguments,
screen and event value)"
  (let ( (mlist (assq 'move on-event::do-list)) )
    (push (cons (get-screen screen) action) (cdr mlist))
  )
)
;;; --------------------------------------------------------------------------
;;; Need to fix the minibuffer properties also, but we have to set them,
;;; and then unmap/map the minibuffer to get the WM to notice the change.
(on-unmap-do 0 'map-screen)
(unmap-screen 0)
(set-property xa-wm-protocols wm-protocol-list 0)
(map-screen 0)				;try for an immediate mapping
;;; --------------------------------------------------------------------------
;;; default key bindings

(global-unset-key "\C-z")
(global-set-key "\C-zo" 'switch-screen)
(global-set-key "\C-z2" 'duplicate-screen)
(global-set-key "\C-z4\C-f" 'find-file-other-screen)
(global-set-key "\C-z4f" 'find-file-other-screen)
(global-set-key "\C-z4b" 'find-buffer-other-screen)
(global-set-key "\C-z0" (definteractive (remove-screen)))
(global-set-key "\C-zm" (definteractive (raise-screen (minibuf-screen))))
(global-set-key "\C-zr" (definteractive (raise-screen)))
(global-set-key "\C-zl" (definteractive (lower-screen)))
(global-set-key "\C-zf" 'focus-current-screen)
(global-set-key "\C-zs" 'switch-screen-noselect)
(global-set-key "\C-zp" 'prev-switch-screen)
(global-set-key "\C-zi" (definteractive (iconify-screen)))
(global-set-key "\C-ze" 'display-event-status)
(global-set-key "\C-za" (definteractive (title (title)) (icon-name (icon-name))))
;;;
(sit-for 1)				;allow various events to go through
